import { memo, useState, useMemo, useEffect } from 'react';
import { NodeAttribute, RelationNodeAttributes, SchemaReactflowRelationNode, toHandleId } from '../../../model';
import { useAppDispatch, useQuerybuilderGraph, useQuerybuilderSettings } from '@graphpolaris/shared/lib/data-access';
import { addWarning } from '@graphpolaris/shared/lib/data-access/store/configSlice';
import { setQuerybuilderGraphology, toQuerybuilderGraphology } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { PillDropdown } from '../../pilldropdown/PillDropdown';
import { Button, DropdownContainer, DropdownTrigger, RelationPill } from '@graphpolaris/shared/lib/components';
import { Handle, Position } from 'reactflow';
import { ArrowDropUp, ArrowDropDown } from '@mui/icons-material';
import { pillWidth } from '@graphpolaris/shared/lib/components/pills/pill.const';

export const QueryRelationPill = memo((node: SchemaReactflowRelationNode) => {
  const data = node.data;
  const graph = useQuerybuilderGraph();
  const graphologyGraph = useMemo(() => toQuerybuilderGraphology(graph), [graph]);
  const settings = useQuerybuilderSettings();
  const dispatch = useAppDispatch();
  const graphologyNodeAttributes = useMemo<RelationNodeAttributes | undefined>(
    () => (graphologyGraph.hasNode(node.id) ? { ...(graphologyGraph.getNodeAttributes(node.id) as RelationNodeAttributes) } : undefined),
    [node.id],
  );
  const attributeEdges = useMemo(
    () => graph.edges.filter((edge) => edge.source === node.id && !!edge?.attributes?.sourceHandleData.attributeType),
    [graph],
  );
  const [openDropdown, setOpenDropdown] = useState(false);

  const [depth, setDepth] = useState<{ min: number; max: number }>({
    min: data.depth.min || settings.depth.min,
    max: data.depth.max || settings.depth.max,
  });

  // TODO: must do this once design is chosen
  // const [direction, setDirection] = useState<RelationshipHandleArrowType>('right');

  useEffect(() => {
    setDepth({ min: data.depth.min || settings.depth.min, max: data.depth.max || settings.depth.max });
  }, [data.depth]);

  const onNodeUpdated = () => {
    if (depth.min < 0) {
      dispatch(addWarning('The minimum depth cannot be smaller than 0'));
    } else if (depth.max > 99) {
      dispatch(addWarning('The maximum depth cannot be larger than 99'));
    } else if (depth.min > depth.max) {
      dispatch(addWarning('The minimum depth cannot be larger than the maximum depth'));
    } else {
      graphologyGraph.setNodeAttribute<any>(node.id, 'depth', depth);
      dispatch(setQuerybuilderGraphology(graphologyGraph));
    }
  };

  // TODO: must do this once design is chosen
  // const onChangeDirection = () => {
  //   if (direction === 'right') {
  //     setDirection('both');
  //     graphologyGraph.setNodeAttribute<any>(node.id, 'direction', 'both');
  //   } else {
  //     setDirection('right');
  //     graphologyGraph.setNodeAttribute<any>(node.id, 'direction', 'right');
  //   }
  //   dispatch(setQuerybuilderGraphology(graphologyGraph));
  // };

  const calcWidth = (data: number) => {
    return data.toString().length + 0.5 + 'ch';
  };

  return (
    <div className="w-fit h-fit p-3 bg-transparent nowheel">
      <RelationPill
        title={
          <div className="flex flex-row w-full">
            <span className="flex-grow text-justify truncate">{data?.name}</span>
            <Button
              variantType="secondary"
              variant="ghost"
              size="2xs"
              iconComponent={openDropdown ? <ArrowDropUp /> : <ArrowDropDown />}
              onMouseDownCapture={(e) => {
                e.stopPropagation();
                e.preventDefault();
                setOpenDropdown(!openDropdown);
              }}
              className={openDropdown ? 'border-secondary-200' : ''}
            />
            {/* <span className="pr-1">
                  <span> [</span>
                  <input
                    className={
                      'bg-inherit text-center appearance-none mx-0.1 rounded-sm ' +
                      (depth.min < 0 || depth.min > depth.max ? ' bg-danger-400 ' : '')
                    }
                    style={{ maxWidth: calcWidth(depth.min) }}
                    type="number"
                    min={0}
                    placeholder={'?'}
                    value={depth.min}
                    onChange={(e) => {
                      setDepth({ ...depth, min: parseInt(e.target.value) });
                    }}
                    onBlur={(e) => {
                      onNodeUpdated();
                    }}
                    onKeyDown={(e) => {
                      if (e.key === 'Enter') {
                        onNodeUpdated();
                      }
                    }}
                  ></input>
                  <span>..</span>
                  <input
                    className={
                      'bg-inherit text-center appearance-none mx-0.1 rounded-sm ' +
                      (depth.max > 99 || depth.min > depth.max ? ' bg-danger-400 ' : '')
                    }
                    style={{ maxWidth: calcWidth(depth.max) }}
                    type="number"
                    min={1}
                    placeholder={'?'}
                    value={depth.max}
                    onChange={(e) => {
                      setDepth({ ...depth, max: parseInt(e.target.value) });
                    }}
                    onBlur={(e) => {
                      onNodeUpdated();
                    }}
                    onKeyDown={(e) => {
                      if (e.key === 'Enter') {
                        onNodeUpdated();
                      }
                    }}
                  ></input>
                  <span>]</span>
                </span> */}
          </div>
        }
        withHandles="horizontal"
        handleLeft={
          <Handle
            // onDoubleClick={onChangeDirection}
            id={toHandleId(data.leftEntityHandleId)}
            position={Position.Left}
            className={'!rounded-none !bg-transparent !w-full !h-full !border-0 !b-0 !right-0 !left-0'}
            type="target"
          ></Handle>
        }
        handleRight={
          <Handle
            // onDoubleClick={onChangeDirection}
            id={toHandleId(data.rightEntityHandleId)}
            position={Position.Right}
            className={'!rounded-none !bg-transparent !w-full !h-full !border-0 !right-0 !left-0'}
            type="source"
          ></Handle>
        }
      >
        {data?.attributes && (
          <PillDropdown
            node={node}
            attributes={data.attributes}
            attributeEdges={attributeEdges.map((edge) => edge?.attributes)}
            open={openDropdown}
            mr={-pillWidth * 0.05}
          />
        )}
      </RelationPill>
    </div>
  );
});

QueryRelationPill.displayName = 'RelationPill';
export default QueryRelationPill;
