import { Edge, GraphQueryResult, Node, useML, useSearchResultData } from '@graphpolaris/shared/lib/data-access';
import { dataColors, visualizationColors } from 'config';
import { Viewport } from 'pixi-viewport';
import {
  Application,
  ColorSource,
  Container,
  DisplayObject,
  FederatedPointerEvent,
  Graphics,
  IPointData,
  Point,
  Sprite,
  Text,
  Texture,
} from 'pixi.js';
import { useEffect, useMemo, useRef, useState } from 'react';
import { LinkType, NodeType } from '../types';
import { NLPopup } from './MatrixPopup';

import { Actions, Interpolations } from 'pixi-actions';

import Color from 'color';
import { createColumn } from './ColumnGraphicsComponent';
import { ReorderingManager } from './ReorderingManager';
import { VisualizationSettingsType } from '../../../common';
import { range, scaleLinear, scaleOrdinal, schemeCategory10 } from 'd3';
import { MatrixVisProps } from '../matrixvis';

type Props = {
  // onClick: (node: NodeType, pos: IPointData) => void;
  // onHover: (data: { node: NodeType; pos: IPointData }) => void;
  // onUnHover: (data: { node: NodeType; pos: IPointData }) => void;
  highlightNodes: NodeType[];
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  graph?: GraphQueryResult;
  settings: MatrixVisProps;
};

const columnsContainer = new Container();
const columnAxisContainer = new Container();
const rowsContainer = new Container();

//////////////////
// MAIN COMPONENT
//////////////////

export const MatrixPixi = (props: Props) => {
  const app = useMemo(() => new Application({ background: 0xffffff, antialias: true, autoDensity: true, eventMode: 'auto' }), []);

  let config = {
    textOffsetX: 100,
    textOffsetY: 100,
    visMapping: [] as any[], // TODO type

    cellHeight: 100,
    cellWidth: 100,
  };

  let columnOrder: string[] = [];
  let rowOrder: string[] = [];

  const [quickPopup, setQuickPopup] = useState<{ node: NodeType; pos: IPointData } | undefined>();
  const [popups, setPopups] = useState<{ node: NodeType; pos: IPointData }[]>([]);
  // const [columnOrder, setColumnOrder] = useState<string[]>([]);

  const nodeMap = useRef(new Map<string, Graphics>());
  const linkMap = useRef(new Map<string, Graphics>());
  const viewport = useRef<Viewport>();
  const ref = useRef<HTMLDivElement>(null);
  const isSetup = useRef(false);
  const ml = useML();
  const searchResults = useSearchResultData();

  // const imperative = useRef<any>(null);

  function resize() {
    const width = ref?.current?.clientWidth || 1000;
    const height = ref?.current?.clientHeight || 1000;
    app.renderer.resize(width, height);
    if (viewport.current) {
      viewport.current.screenWidth = width;
      viewport.current.worldWidth = width;
      viewport.current.worldHeight = height;
      viewport.current.screenHeight = height;
    }

    if (props.graph) {
      setup();
    }

    app.render();
  }

  useEffect(() => {
    if (!ref.current) return;
    const resizeObserver = new ResizeObserver(() => {
      resize();
    });
    resizeObserver.observe(ref.current);
    return () => resizeObserver.disconnect(); // clean up
  }, []);

  useEffect(() => {
    if (ref.current && ref.current.children.length === 0) {
      ref.current.appendChild(app.view as HTMLCanvasElement);
      resize();
    }
  }, [ref]);

  useEffect(() => {
    // console.log('graph changed', props.graph, ref.current, ref.current.children.length > 0, imperative.current);
    if (props.graph && ref.current && ref.current.children.length > 0) {
      if (!isSetup.current) setup();
      else update();
    }
  }, [props.graph]);

  useEffect(() => {
    // console.log('graph changed', props.graph, ref.current, ref.current.children.length > 0, imperative.current);
    if (props.graph && ref.current && ref.current.children.length > 0) {
      setup();
    }
  }, [props.settings]);

  // TODO implement search results
  // useEffect(() => {
  //   if (props.graph) {
  //     props.graph.nodes.forEach((node: NodeType) => {
  //       const gfx = nodeMap.current.get(node.id);
  //       if (!gfx) return;
  //       const isNodeInSearchResults = searchResults.nodes.some((resultNode) => resultNode.id === node.id);
  //       gfx.alpha = isNodeInSearchResults || searchResults.nodes.length === 0 ? 1 : 0.05;
  //     });

  //     props.graph.links.forEach((link: LinkType) => {
  //       const gfx = linkMap.current.get(link.id);
  //       if (!gfx) return;
  //       const isLinkInSearchResults = searchResults.edges.some((resultEdge) => resultEdge.id === link.id);
  //       gfx.alpha = isLinkInSearchResults || searchResults.edges.length === 0 ? 1 : 0.05;
  //     });
  //   }
  // }, [searchResults]);

  function onButtonDown(event: FederatedPointerEvent) {
    console.log(
      event.currentTarget,
      // graph.nodes.find((node) => node.id === event.currentTarget.name)
    );
  }

  const update = (forceClear = false) => {
    setPopups([]);
    if (!props.graph || !ref.current) return;

    if (props.graph) {
      if (forceClear) {
        nodeMap.current.clear();
        linkMap.current.clear();
        rowsContainer.removeChildren();
        columnsContainer.removeChildren();
      }
    }
  };

  const onHoverColumn = (event: any, currentNode: Node | null) => {
    app.renderer.plugins.interaction.cursorStyles.pointer = 'crosshair';
    if (!currentNode) {
      rowsContainer.children.forEach((row) => {
        row.alpha = 1;
      });
      columnsContainer.children.forEach((col) => {
        col.alpha = 1;
      });
      columnAxisContainer.children.forEach((colText) => {
        colText.alpha = 1;
      });

      return;
    }

    columnsContainer.children.forEach((col) => {
      col.alpha = 0.25;
    });
    event.currentTarget.alpha = 1;

    columnAxisContainer.children.forEach((colText) => {
      colText.alpha = 0.25;
      if (colText.name?.endsWith(currentNode._id)) colText.alpha = 1;
    });

    if (!props.graph) return;
    rowsContainer.children.forEach((col) => {
      col.alpha = 0.25;
    });

    if (!props.graph) return;
    rowsContainer.children.forEach((col) => {
      col.alpha = 0.25;
    });

    const edgesForThisColumn = props.graph.edges.filter((edge) => {
      return edge.from === currentNode._id || edge.to === currentNode._id;
    });

    edgesForThisColumn.forEach((edge) => {
      let row = rowsContainer.getChildByName(edge.to);
      if (row) row.alpha = 1;

      row = rowsContainer.getChildByName(edge.from);
      if (row) row.alpha = 1;
    });
  };

  const reorderColumns = (columnOrder: string[]) => {
    const columns = columnsContainer.children;
    const columnsText = columnAxisContainer.children;
    let columnPositions: Point[] = [];
    for (let i = 0; i < columns.length; i++) {
      const col = columns[i];
      columnPositions.push(new Point(col.position.x, col.position.y));
    }

    let columnTextPositions: Point[] = [];
    for (let i = 0; i < columnsText.length; i++) {
      const col = columnsText[i];
      columnTextPositions.push(new Point(col.position.x, col.position.y));
    }

    if (columnTextPositions.length !== columnPositions.length)
      throw new Error(
        'columnTextPositions and columnPositions have different length ' + columnTextPositions.length + ' / ' + columnPositions.length,
      );

    for (let i = 0; i < columns.length; i++) {
      const column = columns[i];
      if (!column.name) throw new Error('column has no name');
      const index = columnOrder.indexOf(column.name);
      if (index === -1) throw new Error('column not found in columnOrder');

      const newPosition = columnPositions[index];
      // move column to new position
      Actions.moveTo(columns[i], newPosition.x, newPosition.y, 1, Interpolations.pow2out).play();

      const newPositionText = columnTextPositions[index];
      // move column to new position
      Actions.moveTo(columnsText[i], newPositionText.x, newPositionText.y, 1, Interpolations.pow2out).play();
    }
  };

  const reorderRows = (
    nodesCol: Node[],
    nodesRow: Node[],
    edges: Edge[],
    rowOrder: string[],
    colorScale: any,
    cellWidth: number,
    cellHeight: number,
  ) => {
    const rows = rowsContainer.children;
    let rowPositions: Point[] = [];
    for (let i = 0; i < rows.length; i++) {
      const row = rows[i];
      rowPositions.push(new Point(row.position.x, row.position.y));
    }

    for (let i = 0; i < rows.length; i++) {
      const row = rows[i] as DisplayObject;
      if (!row.name) throw new Error('row has no name');
      const index = rowOrder.indexOf(row.name);
      if (index === -1) throw new Error('row not found in rowOrder');

      const newPosition = rowPositions[index];
      // move column to new position
      Actions.moveTo(row, newPosition.x, newPosition.y, 1, Interpolations.pow2out).play();
    }

    const colOrder = columnsContainer.children.map((col) => col.name) as string[];
    if (!colOrder) throw new Error('colOrder is undefined');
    columnsContainer.removeChildren();
    setupColumns(edges, colOrder, rowOrder);
  };

  const setup = () => {
    if (!props.graph) throw Error('Graph is undefined; setup not possible');
    columnsContainer.removeChildren();
    rowsContainer.removeChildren();
    columnAxisContainer.removeChildren();
    app.stage.removeChildren();

    const size = ref.current?.getBoundingClientRect();
    viewport.current = new Viewport({
      screenWidth: size?.width || 1000,
      screenHeight: size?.height || 1000,
      worldWidth: size?.width || 1000,
      worldHeight: size?.height || 1000,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });

    app.stage.addChild(viewport.current);

    viewport.current.addChild(columnsContainer);
    viewport.current.addChild(rowsContainer);

    const groupByType = props.graph.nodes.reduce(
      (group: any, node: Node) => {
        if (!group[node.label]) group[node.label] = [];
        group[node.label].push(node);
        return group;
      },
      {} as { [key: string]: Node[] },
    );

    // order groupByType by size
    const ordered = Object.entries(groupByType).sort((a: any[], b: any[]) => b[1].length - a[1].length);

    let cols = [] as Node[];
    let rows = [] as Node[];
    if (ordered.length == 2) {
      cols = ordered[0][1] as Node[];
      rows = ordered[1][1] as Node[];
    } else if (ordered.length == 1) {
      cols = ordered[0][1] as Node[];
      rows = ordered[0][1] as Node[];
    } else {
      // show text that there are no nodes on the viewport
      const finalTextSize = 25;
      const bigRandomScaleFactor = 10;
      const basicText = new Text('Result is empty', { fontSize: finalTextSize * bigRandomScaleFactor, fill: 0x000000, align: 'center' });
      basicText.position.set(viewport.current?.screenWidth / 2 || 1000, viewport.current?.screenHeight / 2 || 1000);
      basicText.scale.set(finalTextSize / (bigRandomScaleFactor * bigRandomScaleFactor));
      basicText.eventMode = 'none';
      viewport.current.addChild(basicText);
      return;
    }

    columnOrder = range(0, cols.length).map((d) => cols[d]._id);
    rowOrder = range(0, rows.length).map((d) => rows[d]._id);
    const reorderingManager = new ReorderingManager(props.graph);
    const newOrdering = reorderingManager.reorderMatrix('count', columnOrder, rowOrder);
    columnOrder = newOrdering.columnOrder;
    rowOrder = newOrdering.rowOrder;

    config.cellWidth = Math.max((size?.width || 1000) / props.graph.nodes.length, (size?.height || 1000) / props.graph.nodes.length);
    config.cellHeight = config.cellWidth;

    setupVisualizationEncodingMapping(props.settings);

    setupColumns(props.graph.edges, columnOrder, rowOrder);
    setupColumnLegend(columnOrder);
    setupColumnInteractivity(cols);

    setupRowLegend(rows, rowOrder);
    setupRowInteractivity(cols, rows, rowOrder, scaleOrdinal(schemeCategory10));

    console.debug('setup matrixvis with graph:', props.graph);

    // activate plugins
    viewport.current.drag().pinch().wheel().animate({}).decelerate({ friction: 0.75 });

    app.ticker.add(tick);
    update();
    isSetup.current = true;
  };

  const setupVisualizationEncodingMapping = (settings: MatrixVisProps) => {
    if (!props.graph) throw new Error('Graph is undefined; cannot setup matrix');
    const visMapping = []; // TODO type

    let colorNeutralString = dataColors.neutral['5'];
    const colorNeutral = Color(colorNeutralString);

    // make adjacency
    // const adjacenyScale = scaleLinear([dataColors.neutral['5'], tailwindColors.entity.DEFAULT]);
    const adjacenyScale = scaleLinear([colorNeutral, visualizationColors.GPSelected.colors[1][0]]);
    visMapping.push({
      attribute: 'adjacency',
      encoding: settings.marks,
      colorScale: adjacenyScale,
      renderFunction: function (i: number, color: ColorSource, gfxContext: Graphics) {
        gfxContext.beginFill(color, 1);

        // TODO get
        // gfx.beginFill(colorScale(node.type)));

        gfxContext.lineStyle(2, 0xffffff);
        if (this.encoding === 'rect') {
          gfxContext.drawRect(0, i * config.cellHeight, config.cellWidth, config.cellHeight);
        }
        if (this.encoding === 'circle') {
          gfxContext.drawCircle(config.cellWidth / 2, i * config.cellHeight + config.cellHeight / 2, config.cellWidth / 2);
        }
        gfxContext.endFill();

        return gfxContext;
      },
    });

    // make label
    // const labels = props.graph.edges.map((edge: Edge) => {
    //   return edge.label;
    // });
    // const tailwindColorsArray = Object.values(tailwindColors.entity).slice(0, labels.length);
    // const scale = scaleOrdinal(tailwindColorsArray.reverse());
    // scale.domain(Array.from(labels));
    // console.log('labels', labels, props.graph.edges);
    // visMapping.push({ attribute: 'label', encoding: 'rect', colorScale: scale });

    // make value
    // const values = props.graph.edges.map((edge: Edge) => {
    //   return edge?.attributes?.Distance as number;
    // });

    // const extent = extent(values);
    // if (extent[0] === undefined || extent[1] === undefined) throw new Error('Extent is undefined');
    // console.log('labels', values, extent, props.graph.edges);

    // const scale = scaleLinear(['red', 'blue']);
    // scale.domain(extent as [number, number]);
    // visMapping.push({ attribute: 'attributes.Distance', encoding: 'rect', colorScale: scale });

    config.visMapping = visMapping;
  };

  function setupColumns(edges: Edge[], columnOrder: string[], rowOrder: string[]) {
    const visMapping = config.visMapping[0];

    if (!visMapping) throw new Error('Cannot setup matrix without visMapping');

    for (let j = 0; j < columnOrder.length; j++) {
      const oneColumn = new Container();
      oneColumn.name = columnOrder[j];

      const edgesForThisColumn = edges.filter((edge) => {
        return edge.from === oneColumn.name || edge.to === oneColumn.name;
      });

      const col = createColumn(rowOrder, edgesForThisColumn, config.visMapping, config.cellWidth, config.cellHeight);
      oneColumn.addChild(col);
      oneColumn.position.set(j * config.cellWidth + config.textOffsetX, config.textOffsetY);

      columnsContainer.addChild(oneColumn);
      oneColumn.alpha = 0;
      Actions.sequence(Actions.delay(j * 0.005 * Math.random()), Actions.fadeIn(oneColumn, 0.8, Interpolations.pow2out)).play();
    }
  }

  const setupColumnInteractivity = (cols: Node[]) => {
    for (let j = 0; j < columnsContainer.children.length; j++) {
      const oneColumn = columnsContainer.children[j];
      oneColumn.interactive = true;
      oneColumn.on('pointerdown', onButtonDown);
      oneColumn.on('pointerover', (event) => {
        const col = cols.find((col) => col._id === oneColumn.name);
        if (col) onHoverColumn(event, col);
      });
      oneColumn.on('pointerout', (event) => {
        onHoverColumn(event, null as any);
      });
    }
  };

  const setupColumnLegend = (columnOrder: string[]) => {
    // columnAxisContainer.removeChildren();
    columnAxisContainer.position.set(config.textOffsetX, 0);
    for (let j = 0; j < columnOrder.length; j++) {
      // from the PixiJS documention: Setting a text object's scale to > 1.0 will result in blurry/pixely display,
      // because the text is not re-rendered at the higher resolution needed to look sharp -
      // it's still the same resolution it was when generated. To deal with this, you can render at a higher
      // initial size and down-scale, instead.
      const finalTextSize = 10;
      const bigRandomScaleFactor = 10;
      const basicText = new Text(columnOrder[j], { fontSize: finalTextSize * bigRandomScaleFactor, fill: 0x000000, align: 'center' });
      basicText.position.set(j * config.cellWidth + config.cellWidth / 4, config.textOffsetY - 5);
      basicText.scale.set(finalTextSize / (bigRandomScaleFactor * bigRandomScaleFactor));
      basicText.rotation = (Math.PI * 3) / 2;
      basicText.eventMode = 'none';
      basicText.name = 'Text_' + columnOrder[j];
      columnAxisContainer.addChild(basicText);
    }
    viewport.current?.addChild(columnAxisContainer);

    const columnAxis = new Sprite(Texture.WHITE);
    columnAxis.width = columnOrder.length * config.cellWidth;
    columnAxis.height = config.textOffsetY;
    columnAxis.tint = 0xaaaaaa;
    columnAxis.position.set(config.textOffsetX, 0);
    columnAxis.alpha = 0.0;
    columnAxis.interactive = true;
    columnAxis.on('pointerdown', (event) => {
      // const reordering = new ReorderColumnsAction(columnOrder);

      if (!props.graph) throw new Error('Graph is undefined; cannot reorder matrix');

      const reorderingManager = new ReorderingManager(props.graph);
      const newOrdering = reorderingManager.reorderMatrix('count', columnOrder, rowOrder);
      columnOrder = newOrdering.columnOrder.reverse();

      reorderColumns(columnOrder);
    });
    // columnAxis.on('pointerdown', reorderColumns);
    columnAxis.on('pointerover', (event) => {
      Actions.fadeTo(columnAxis, 0.5, 0.5, Interpolations.pow2out).play();
    });
    columnAxis.on('pointerout', (event) => {
      Actions.fadeTo(columnAxis, 0.0, 0.5, Interpolations.pow2out).play();
    });
    viewport.current?.addChild(columnAxis);
  };

  function shuffle(array: any[]) {
    var m = array.length,
      t,
      i;

    // While there remain elements to shuffle…
    while (m) {
      // Pick a remaining element…
      i = Math.floor(Math.random() * m--);

      // And swap it with the current element.
      t = array[m];
      array[m] = array[i];
      array[i] = t;
    }

    return array;
  }

  function setupRowLegend(rows: Node[], rowOrder: string[]) {
    for (let i = 0; i < rowOrder.length; i++) {
      const rowID = rowOrder[i];
      const finalTextSize = 10;
      const bigRandomScaleFactor = 10;
      const basicText = new Text(rowID, { fontSize: finalTextSize * bigRandomScaleFactor, fill: 0x000000, align: 'right' });
      basicText.anchor.x = 1;
      basicText.anchor.y = 0;
      basicText.scale.set(finalTextSize / (bigRandomScaleFactor * bigRandomScaleFactor));
      basicText.position.set(config.textOffsetX - 5, i * config.cellHeight + config.textOffsetY + config.cellHeight / 4);
      basicText.name = rowID;
      rowsContainer.addChild(basicText);

      Actions.sequence(Actions.delay(i * 0.005 * Math.random()), Actions.fadeIn(basicText, 0.5, Interpolations.pow2out)).play();
    }
  }

  const setupRowInteractivity = (cols: Node[], rows: Node[], rowOrder: string[], colorScale: any) => {
    const rowAxis = new Sprite(Texture.WHITE);
    rowAxis.width = config.textOffsetX;
    rowAxis.height = rows.length * config.cellHeight;
    rowAxis.tint = 0xaaaaaa;
    rowAxis.position.set(0, config.textOffsetY);
    rowAxis.alpha = 0.0;
    rowAxis.interactive = true;
    rowAxis.on('pointerdown', (event) => {
      if (props.graph) {
        rowOrder = shuffle(rowOrder);
        reorderRows(cols, rows, props?.graph.edges, rowOrder, colorScale, config.cellWidth, config.cellHeight);
      }
    });
    rowAxis.on('pointerover', (event) => {
      Actions.fadeTo(rowAxis, 0.5, 0.5, Interpolations.pow2out).play();
    });
    rowAxis.on('pointerout', (event) => {
      Actions.fadeTo(rowAxis, 0.0, 0.5, Interpolations.pow2out).play();
    });
    viewport.current?.addChild(rowAxis);
  };

  const tick = (delta: number) => {
    if (props.graph) {
      Actions.tick(delta / 60);
    }
  };

  return (
    <>
      {popups.map((popup) => (
        <NLPopup onClose={() => {}} data={popup} key={popup.node.id} />
      ))}
      {quickPopup && <NLPopup onClose={() => {}} data={quickPopup} />}
      <div className="h-full w-full overflow-hidden" ref={ref}></div>
    </>
  );
};
